_npcs = {}

--------------------------------------------------------------
--  registerNPC(id)
--------------------------------------------------------------
--
--  Register an existing NPC to the mdNPC framework.
--  WARNING: Only NPC's of the type md_XXX can be registered
--
--------------------------------------------------------------
--  Parameters:
--   id:   unique ID of the NPC to be registered
--------------------------------------------------------------
--  Example: mdNPC.script:registerNPC("myNPC")
--------------------------------------------------------------
function registerNPC(self, id)
	local npc = findEntity(id)
	if npc ~= nil then
		npc.move:setCooldown(0)
		if id ~= nil then
			local npcData = {
				id = id,
				level = npc.level,
				x = npc.x,
				y = npc.y,
				e = npc.e,
				facing = npc.facing,
				route = "",
				repeatRoute = false,
				routeIndex = 0,
				invulnerable = true,
				tag = "",
				onClick = nil,
				onGive = nil
			}
			table.insert(_npcs, npcData)
		end
	else
		print("NPC not found: ", id)
	end
end

--------------------------------------------------------------
--  isWalking(id)
--------------------------------------------------------------
--
--  Indicates if the npc is currently walking or not
--
--------------------------------------------------------------
--  Parameters:
--   id:           unique ID of NPC
--------------------------------------------------------------
--  Return values:
--   isWalking	TRUE if the npc is walking, false if he is not
--------------------------------------------------------------
function isWalking(self, id)
	local npc = getOrRegisterNPC(id)
	if npc ~= nil then
		return npc.route ~= "" and (npc.routeIndex <= string.len(npc.route))
	end
end

--------------------------------------------------------------
--  setOnInteract(functionName)
--------------------------------------------------------------
--
--  Defines what function to call when the user interacts with
--  the NPC. 
--
--  WARNING: The function defined here will only be called
--           if the NPC is not walking.
--
------------------------------------------------------------------------------
--  Parameters:
--   id:   unique ID of the NPC to be registered
--   functionName:   function to call, without parenthesis.
------------------------------------------------------------------------------
--  Example: mdNPC.script:setOnInteract("myNPC", myScript.script.myNPCOnClick)
------------------------------------------------------------------------------
function setOnInteract(self, id, onClickFunction)
	local npc = getOrRegisterNPC(id)
	if npc ~= nil then
		npc.onClick = onClickFunction
	end
end

--------------------------------------------------------------
--  setOnGive(functionName)
--------------------------------------------------------------
--
--  Defines what function to call when the user clicks on 
--  the NPC while they have an item on the cursor.
--
--  WARNING: The function defined here will only be called
--           if the NPC is not walking
--
--
--   This function should return TRUE if the item on the 
--   mouseCursor needs to be destroyed, or FALSE to leave
--   the item there.
--
--------------------------------------------------------------
--  Parameters:
--   id:   unique ID of the NPC to be registered
--   functionName:   function to call, without parenthesis.
--------------------------------------------------------------------------
--  Example: mdNPC.script:setOnGive("myNPC", myScript.script.myNPCOnClick)
--------------------------------------------------------------------------
function setOnGive(self, id, onGiveFunction)
	local npc = getOrRegisterNPC(id)
	if npc ~= nil then
		npc.onGive = onGiveFunction
	end
end

--------------------------------------------------------------
--  setRoute(id, route, repeatRoute)
--------------------------------------------------------------
--
--  Lets the npc walk the indicated route.
--
--------------------------------------------------------------
--  Parameters:
--   id:     ID of npc
--   route:  string with chain of commands of route to walk
--           (f)orward, (l)eft, (r)ight, or (w)ait
--           Example: "lffr" -> left, forward, forward, right
--   repeat: Indicates if the route needs to be repeated after
--           it has been fully processed. Default = false
--------------------------------------------------------------
function setRoute(self, id, route, repeatRoute)
	local npc = getOrRegisterNPC(id)
	if npc ~= nil then
		npc.route = route.."w"
		if repeatRoute == nil then
			npc.repeatRoute = false
		else
			npc.repeatRoute = repeatRoute
		end
		npc.routeIndex = 1
	end
end




--------------------------------------------------------------
--  appendRoute(id, route)
--------------------------------------------------------------
--
--  Appends the given route to the npcs current route. If the
--  npc is not walking, he will start to walk.
--
--------------------------------------------------------------
--  Parameters:
--   id:     unique ID of npc returned by registerNPC()
--   route:  string with chain of commands of route to walk
--           (f)orward, (l)eft, (r)ight, or (w)ait
--           Example: "lffr" -> left, forward, forward, right
--------------------------------------------------------------
function appendRoute(self, id, route)
	local npc = getOrRegisterNPC(id)
	if npc ~= nil then
		if isWalking(self, id) then
			npc.route = npc.route..route
		else
			setRoute(self, id, route, false)
		end		
	end
end

-----------------------------------------------------------
--
-- INTERNAL FUNCTIONS
--
-----------------------------------------------------------
function onThink(self, npc)
	local npcData = getNPC(npc.id)
	if npcData ~= nil then

		if npcData.route ~= "" and npcData.routeIndex <= string.len(npcData.route) then
		
			local didMove = false
			local cmd = string.sub(npcData.route, npcData.routeIndex, npcData.routeIndex)
			
			if cmd == "l" then
				npc.brain:turnLeft()
				didMove = true
			elseif cmd == "r" then
				npc.brain:turnRight()
				didMove = true
			elseif cmd == "f" then
				didMove = npc.brain:moveForward()
			elseif cmd == "w" then
				npc.brain:wait()
				didMove = true
			end

			if didMove then
				npcData.routeIndex = npcData.routeIndex + 1
				if npcData.routeIndex > string.len(npcData.route) then
					if npcData.repeatRoute == true then
						npcData.routeIndex = 1
					end
				end
			else
				--npc.brain:wait()
			end
		else
			-- Done walking, just wait
			npc.brain:wait()
		end
	else
		npc.brain:wait()
	end
end

function onClick(self, npc)
	local npcData = getNPC(npc.id)
	if npcData ~= nil then
		if not isWalking(self, npc.id) then
			if getMouseItem() == nil then
				if npcData.onClick ~= nil then
					faceParty(self, npc)
					local turn2Player = npcData.onClick(self, npc)
				end
			else
				if npcData.onGive ~= nil then
					faceParty(self, npc)
					local destroyItem = npcData.onGive(self, npc, getMouseItem())
					if destroyItem == nil or destroyItem == true then
						setMouseItem(nil)
					end
				end
			end
		end
	else
		print("No interact/give function definined for npc:", npc.id)
	end
end

function faceParty(self, npc)	
	local d = npc.facing - party.facing
	if d == -3 or d == 1 then setRoute(self, npc.id, "r", false) end
	if d == -1 or d == 3 then setRoute(self, npc.id, "l", false) end
	if d == 0 then setRoute(self, npc.id, "ll", false) end
end

function getOrRegisterNPC(id)
	local npc = getNPC(id)
	if npc == nil then
		registerNPC(self, id)
		npc = getNPC(id)
	end
	return npc
end

function getNPC(id)
	for _, npc in pairs(_npcs) do
		if npc.id == id then
			return npc
		end
	end
end
